/*
 * MIDTacToe: tic-tac-toe for modile phones
 * Copyright (C) 2003  Jason Pratt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jasonpratt.midtactoe;

import javax.microedition.lcdui.*;

/**
 * <p>Copyright: Copyright (c) 2003</p>
 * @author Jason Pratt
 */
public class TicTacToeCanvas extends Canvas {
	public static final Command EXIT_COMMAND = new Command( "Exit", Command.EXIT, 5 );
	public static final Command INTERRUPT_THINKING_COMMAND = new Command( "Stop", Command.STOP, 5 );
	public static final Command NEW_GAME_COMMAND = new Command( "New", Command.SCREEN, 5 );

	public static int UP = 1;
	public static int DOWN = 2;
	public static int LEFT = 3;
	public static int RIGHT = 4;

	public static int STATUS_NULL = 0;
	public static int STATUS_LEGAL = 1;
	public static int STATUS_ILLEGAL = 2;

	protected int screenWidth;
	protected int screenHeight;
	protected Font wordFont;
	protected int wordHeight;
	protected int wordBuffer;
	protected int line1Y;
	protected int line2Y;
	protected int line3Y;
	protected int arrowsHeight2;
	protected int arrowsWidth2;
	protected int arrowsY2;
	protected int arrowsHeight3;
	protected int arrowsWidth3;
	protected int arrowsY3;
	protected int boardBuffer;
	protected int boardSize;
	protected int boardX;
	protected int boardY;
	protected int lineThickness;
	protected int squareBuffer;
	protected int squareSize;
	protected int[] squareX;
	protected int[] squareY;
	protected Font numberFont;
	protected int numberHeight;
	protected int numberXOffset;
	protected int numberYOffset;
	protected boolean isColor;
	protected int numColors;

	protected int backgroundColor;
	protected int numberColor;
	protected int line1Color;
	protected int line2Color;
	protected int line3Color;
	protected int boardColor;
	protected int xColor;
	protected int oColor;
	protected int winColor;
	protected int loseColor;
	protected int winBackgroundColor;
	protected int tieBackgroundColor;
	protected int feedbackColor;
	protected int errorColor;
	protected int arrowColor;

	protected int[] backgroundColors;

	protected int upStatus;
	protected int downStatus;
	protected int leftStatus;
	protected int rightStatus;
	protected int[] numbersStatus;

	protected MIDTacToe midTacToe;
	protected boolean badMoveOccured;
	protected int badMove;
	protected boolean thinking;
	protected boolean interrupted;
	protected Thread thinkingThread;
	protected ComputerPlayRunnable computerPlayRunnable = new ComputerPlayRunnable();

	public TicTacToeCanvas( MIDTacToe midTacToe ) {
		this.midTacToe = midTacToe;
		init();
	}

	private void init() {
		addCommand( EXIT_COMMAND );
		addCommand( NEW_GAME_COMMAND );

		setCommandListener( new MainCommandAdapter() );

		screenWidth = getWidth();
		screenHeight = getHeight();
		wordFont = Font.getFont( Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL );
		wordHeight = wordFont.getHeight();
		wordBuffer = 1;
		line1Y = wordBuffer;
		line3Y = screenHeight - wordBuffer;
		line2Y = line3Y - wordHeight;
		arrowsHeight3 = (wordHeight - 2 - 2)/2;
		arrowsWidth3 = 2*arrowsHeight3 - 1;
		arrowsY3 = screenHeight - wordHeight + 1;
		arrowsHeight2 = arrowsWidth3;
		arrowsWidth2 = arrowsHeight3;
		arrowsY2 = screenHeight - 2*wordHeight + (wordHeight - arrowsHeight2)/2;
		boardBuffer = 2;
		lineThickness = 2;
		boardSize = screenHeight - (2*wordBuffer + 3*wordHeight + 2*boardBuffer);
		if( (boardSize + 2*boardBuffer) > screenWidth ) {
			boardSize = screenWidth - 2*boardBuffer;
		}
		while( ((boardSize - 2*lineThickness) % 3) != 0 ) {
			boardSize--;
		}
		boardX = (screenWidth - boardSize)/2;
		boardY = wordHeight + wordBuffer + boardBuffer;
		squareBuffer = 1;
		squareSize = (boardSize - 2*lineThickness - 6*squareBuffer)/3;
		int offset1 = squareBuffer;
		int offset2 = 3*squareBuffer + lineThickness + squareSize;
		int offset3 = 5*squareBuffer + 2*lineThickness + 2*squareSize;
		squareX = new int[] { boardX + offset1, boardX + offset2, boardX + offset3, boardX + offset1, boardX + offset2, boardX + offset3, boardX + offset1, boardX + offset2, boardX + offset3 };
		squareY = new int[] { boardY + offset1, boardY + offset1, boardY + offset1, boardY + offset2, boardY + offset2, boardY + offset2, boardY + offset3, boardY + offset3, boardY + offset3 };
		numberFont = Font.getFont( Font.FACE_PROPORTIONAL, Font.STYLE_BOLD, Font.SIZE_LARGE );
		numberHeight = numberFont.getBaselinePosition();
		if( numberHeight > squareSize ) {
			numberFont = Font.getFont( Font.FACE_PROPORTIONAL, Font.STYLE_BOLD, Font.SIZE_MEDIUM );
			numberHeight = numberFont.getBaselinePosition();
		}
		if( numberHeight > squareSize ) {
			numberFont = Font.getFont( Font.FACE_PROPORTIONAL, Font.STYLE_BOLD, Font.SIZE_SMALL );
			numberHeight = numberFont.getBaselinePosition();
		}
		numberXOffset = squareSize/2;
		numberYOffset = (squareSize - numberHeight)/2;

		isColor = midTacToe.display.isColor();
		numColors = midTacToe.display.numColors();

		if( isColor ) {
			backgroundColor = 0xFFFFFF;
			numberColor = 0xAAAAAA;
			line1Color = 0x000000;
			line2Color = 0x000000;
			line3Color = 0x000000;
			boardColor = 0x000000;
			xColor = 0x000088;
			oColor = 0x770088;
			winColor = 0x008800;
			loseColor = 0x880000;
			winBackgroundColor = 0xAAFFAA;
			tieBackgroundColor = 0xFFCCCC;
			feedbackColor = 0xFFFF99;
			errorColor = 0xFF0000;
			arrowColor = 0x000000;
		} else if( numColors > 255 ) {
			backgroundColor = 0xFFFFFF;
			numberColor = 0xAAAAAA;
			line1Color = 0x000000;
			line2Color = 0x000000;
			line3Color = 0x000000;
			boardColor = 0x000000;
			xColor = 0x000000;
			oColor = 0x000000;
			winColor = 0x000000;
			loseColor = 0x000000;
			winBackgroundColor = 0xAAAAAA;
			tieBackgroundColor = 0xAAAAAA;
			feedbackColor = 0x808080;
			errorColor = 0x000000;
			arrowColor = 0x000000;
		} else {
			backgroundColor = 0xFFFFFF;
			numberColor = 0x000000;
			line1Color = 0x000000;
			line2Color = 0x000000;
			line3Color = 0x000000;
			boardColor = 0x000000;
			xColor = 0x000000;
			oColor = 0x000000;
			winColor = 0x000000;
			loseColor = 0x000000;
			winBackgroundColor = 0xFFFFFF;
			tieBackgroundColor = 0xFFFFFF;
			feedbackColor = 0xFFFFFF;
			errorColor = 0x000000;
			arrowColor = 0x000000;
		}
		backgroundColors = new int[9];

		numbersStatus = new int[9];
	}

	/////////////
	// painting
	/////////////
	public void paint( Graphics g ) {
		g.setColor( backgroundColor );
		g.fillRect( 0, 0, screenWidth, screenHeight );

		for( int i = 0; i < 9; i++ ) {
			backgroundColors[i] = backgroundColor;
		}

		boolean xWin = TicTacToeEngine.isXWin( midTacToe.currentBoard );
		boolean oWin = TicTacToeEngine.isOWin( midTacToe.currentBoard );
		boolean tie = (! xWin) && (! oWin) && (TicTacToeEngine.getNumMovesLeft( midTacToe.currentBoard ) < 1);

		// status calculations
		String status = "";
		int statusColor = line1Color;
		if( badMoveOccured ) {
			status = "Illegal move!";
			statusColor = errorColor;
		} else if( thinking ) {
			if( midTacToe.currentPlayer == TicTacToeEngine.X_PLAYER ) {
				status = "X's turn: Thinking...";
				statusColor = xColor;
			} else {
				status = "O's turn: Thinking...";
				statusColor = oColor;
			}
		} else if( xWin ) {
			status = "X wins";
			statusColor = xColor;
		} else if( oWin ) {
			status = "O wins";
			statusColor = oColor;
		} else if( tie ) {
			status = "Draw";
		} else if( midTacToe.currentPlayer == TicTacToeEngine.X_PLAYER ) {
			status = "X's turn";
			statusColor = xColor;
		} else {
			status = "O's turn";
			statusColor = oColor;
		}

		// difficulty calculations
		String difficulty;
		if( midTacToe.difficulty == 0 ) {
			difficulty = "Level: 0 Easy";
		} else if( midTacToe.difficulty == 9 ) {
			difficulty = "Level: 9 Hopeless";
		} else {
			difficulty = "Level: " + midTacToe.difficulty;
		}

		// "You are" calculations
		String youAre;
		int youAreColor;
		if( midTacToe.userPlayer == TicTacToeEngine.X_PLAYER ) {
			youAre = "You are: X";
			youAreColor = xColor;
		} else {
			youAre = "You are: O";
			youAreColor = oColor;
		}

		// draw status
		g.setColor( statusColor );
		g.drawString( status, 1, line1Y, Graphics.TOP | Graphics.LEFT );

		// draw arrows
		g.setColor( arrowColor );
		if( midTacToe.difficulty > 0 ) {
			drawTriangle( g, 1, arrowsY2, arrowsHeight2, LEFT );
		}
		if( midTacToe.difficulty < 9 ) {
			drawTriangle( g, 1 + arrowsWidth2 + 2, arrowsY2, arrowsHeight2, RIGHT );
		}
		drawTriangle( g, 1, arrowsY3, arrowsHeight3, UP );
		drawTriangle( g, 1, arrowsY3 + arrowsHeight3 + 2, arrowsHeight3, DOWN );

		//draw difficulty
		g.setColor( line2Color );
		g.drawString( difficulty, 1 + 2*arrowsWidth2 + 2 + 2, line2Y, Graphics.BASELINE | Graphics.LEFT );

		//draw "You are"
		g.setColor( youAreColor );
		g.drawString( youAre, 1 + 2*arrowsWidth2 + 2 + 2, line3Y, Graphics.BASELINE | Graphics.LEFT );

		// draw board lines
		g.setColor( boardColor );
		int x1, y1, x2, y2;
		x1 = x2 = boardX + squareSize + 2*squareBuffer;
		y1 = boardY;
		y2 = boardY + boardSize - 1;
		for( int i = 0; i < lineThickness; i++ ) {
			g.drawLine( x1 + i, y1, x2 + i, y2 );
		}
		x1 = x2 = boardX + 2*squareSize + 4*squareBuffer + lineThickness;
		for( int i = 0; i < lineThickness; i++ ) {
			g.drawLine( x1 + i, y1, x2 + i, y2 );
		}
		x1 = boardX;
		x2 = boardX + boardSize - 1;
		y1 = y2 = boardY + squareSize + 2*squareBuffer;
		for( int i = 0; i < lineThickness; i++ ) {
			g.drawLine( x1, y1 + i, x2, y2 + i );
		}
		y1 = y2 = boardY + 2*squareSize + 4*squareBuffer + lineThickness;
		for( int i = 0; i < lineThickness; i++ ) {
			g.drawLine( x1, y1 + i, x2, y2 + i );
		}

		// draw background
		if( xWin || oWin ) {
			int positions;
			if( xWin ) {
				positions = TicTacToeEngine.getXWinPositions( midTacToe.currentBoard );
			} else {
				positions = TicTacToeEngine.getOWinPositions( midTacToe.currentBoard );
			}
			for( int i = 0; i < 9; i++ ) {
				if( (positions & TicTacToeEngine.allMoves[i]) > 0 ) {
					backgroundColors[i] = winBackgroundColor;
				}
			}
		} else if( tie ) {
			for( int i = 0; i < 9; i++ ) {
				backgroundColors[i] = tieBackgroundColor;
			}
		}
		for( int i = 0; i < 9; i++ ) {
			if( numbersStatus[i] == STATUS_ILLEGAL ) {
				backgroundColors[i] = errorColor;
			} else if( numbersStatus[i] == STATUS_LEGAL ) {
				backgroundColors[i] = feedbackColor;
			}
			if( backgroundColors[i] != backgroundColor ) {
				drawSquare( g, i, backgroundColors[i] );
			}
		}

		// draw board contents
		for( int i = 0; i < 9; i++ ) {
			if( (TicTacToeEngine.xMoves[i] & midTacToe.currentBoard) == TicTacToeEngine.xMoves[i] ) {
				drawX( g, i );
			} else if( (TicTacToeEngine.oMoves[i] & midTacToe.currentBoard) == TicTacToeEngine.oMoves[i] ) {
				drawO( g, i );
			} else {
				drawNumber( g, i );
			}
		}
	}

	public void drawX( Graphics g, int position ) {
		int x = squareX[position];
		int y = squareY[position];

		g.setColor( xColor );

		g.drawLine( x, y, x + squareSize - 1, y + squareSize - 1 );
		g.drawLine( x, y + 1, x + squareSize - 2, y + squareSize - 1 );
		g.drawLine( x + 1, y, x + squareSize - 1, y + squareSize - 2 );

		g.drawLine( x, y + squareSize - 1, x + squareSize - 1, y );
		g.drawLine( x, y + squareSize - 2, x + squareSize - 2, y );
		g.drawLine( x + 1, y + squareSize - 1, x + squareSize - 1, y + 1 );
	}

	public void drawO( Graphics g, int position ) {
		int x = squareX[position];
		int y = squareY[position];
		int thickness = 3;

		g.setColor( oColor );
		g.fillArc( x, y, squareSize, squareSize, 0, 360 );
		g.setColor( backgroundColors[position] );
		g.fillArc( x + thickness, y + thickness, squareSize - 2*thickness, squareSize - 2*thickness, 0, 360 );
	}

	public void drawNumber( Graphics g, int position ) {
		int x = squareX[position];
		int y = squareY[position];

		g.setColor( numberColor );
		g.drawString( Integer.toString( position + 1 ), x + numberXOffset, y + numberYOffset, Graphics.TOP | Graphics.HCENTER );
	}

	public void drawSquare( Graphics g, int position, int color ) {
		int x = squareX[position] - squareBuffer;
		int y = squareY[position] - squareBuffer;
		int size = squareSize + 2*squareBuffer;

		g.setColor( color );
		g.fillRect( x, y, size, size );
	}

	public void drawTriangle( Graphics g, int x, int y, int height, int direction ) {
		if( direction == UP ) {
			int _x, _y;
			int width = (height*2) - 1;
			for( _y = y + height - 1, _x = x; _y >= y; _y--, _x++, width -= 2 ) {
				g.drawLine( _x, _y, _x + width, _y );
			}
		} else if( direction == DOWN ) {
			int _x, _y;
			int width = (height*2) - 1;
			int yMax = y + height;
			for( _y = y, _x = x; _y < yMax; _y++, _x++, width -= 2 ) {
				g.drawLine( _x, _y, _x + width, _y );
			}
		} else if( direction == LEFT ) {
			int _x, _y;
			int _height = height - (1 - (height & 1)); // if even, subtract 1
			int width = (_height + 1)/2;
			for( _x = x + width - 1, _y = y; _x >= x; _x--, _y++, _height -= 2 ) {
				g.drawLine( _x, _y, _x, _y + _height );
			}
		} else if( direction == RIGHT ) {
			int _x, _y;
			int _height = height - (1 - (height & 1)); // if even, subtract 1
			int width = (_height + 1)/2;
			int xMax = x + width;
			for( _x = x, _y = y; _x < xMax; _x++, _y++, _height -= 2 ) {
				g.drawLine( _x, _y, _x, _y + _height );
			}
		}
	}

	/////////////
	// events
	/////////////
	public void newGame() {
		midTacToe.currentBoard = 0;
		midTacToe.currentPlayer = TicTacToeEngine.X_PLAYER;
		thinking = false;
		badMoveOccured = false;
		haveComputerPlayIfAppropriate();
	}

	public void haveComputerPlayIfAppropriate() {
		if( (thinking == false) && (! midTacToe.isGameOver()) && (midTacToe.currentPlayer != midTacToe.userPlayer) ) {
			thinkingThread = new Thread( computerPlayRunnable );
			thinkingThread.start();
		}
	}

	public void makeMove( int position ) {
		badMoveOccured = false;
		if( thinking ) {
			interruptThinking();
		}

		if( TicTacToeEngine.isXWin( midTacToe.currentBoard ) || TicTacToeEngine.isOWin( midTacToe.currentBoard ) || (TicTacToeEngine.getNumMovesLeft( midTacToe.currentBoard ) < 1) ) {
			//TODO: game already over feedback
			numbersStatus[position] = STATUS_ILLEGAL;
		} else if( (midTacToe.currentBoard & TicTacToeEngine.allMoves[position]) != 0 ) {
			//TODO: illegal move feedback
			badMoveOccured = true;
			numbersStatus[position] = STATUS_ILLEGAL;
		} else {
			numbersStatus[position] = STATUS_LEGAL;
			if( midTacToe.currentPlayer == TicTacToeEngine.X_PLAYER ) {
				TicTacToeCanvas.this.midTacToe.currentBoard |= TicTacToeEngine.xMoves[position];
			} else {
				TicTacToeCanvas.this.midTacToe.currentBoard |= TicTacToeEngine.oMoves[position];
			}
			TicTacToeCanvas.this.midTacToe.currentPlayer = 1 - TicTacToeCanvas.this.midTacToe.currentPlayer;
		}
		haveComputerPlayIfAppropriate();
	}

	public void interruptThinking() {
		interrupted = true;
		thinking = false;
		removeCommand( INTERRUPT_THINKING_COMMAND );
		addCommand( NEW_GAME_COMMAND );
		TicTacToeEngine.interruptCalculation();
		repaint();
	}

	public class ComputerPlayRunnable implements Runnable {
		public void run() {
			TicTacToeCanvas.this.thinking = true;
			TicTacToeCanvas.this.removeCommand( NEW_GAME_COMMAND );
			TicTacToeCanvas.this.addCommand( INTERRUPT_THINKING_COMMAND );
			TicTacToeCanvas.this.repaint();

			int depth = Math.min( TicTacToeCanvas.this.midTacToe.difficulty, TicTacToeEngine.getNumMovesLeft( TicTacToeCanvas.this.midTacToe.currentBoard ) );
			int newBoard = TicTacToeEngine.getBestMove( TicTacToeCanvas.this.midTacToe.currentBoard, TicTacToeCanvas.this.midTacToe.currentPlayer, depth );

			if( ! interrupted ) {
				TicTacToeCanvas.this.midTacToe.currentBoard = newBoard;
				TicTacToeCanvas.this.midTacToe.currentPlayer = 1 - TicTacToeCanvas.this.midTacToe.currentPlayer;
				TicTacToeCanvas.this.removeCommand( INTERRUPT_THINKING_COMMAND );
				TicTacToeCanvas.this.addCommand( NEW_GAME_COMMAND );
				TicTacToeCanvas.this.thinking = false;
				TicTacToeCanvas.this.repaint();
			} else {
				interrupted = false;
			}
		}
	}

	public class MainCommandAdapter implements CommandListener {
		public void commandAction( Command command, Displayable displayable ) {
			if( command == TicTacToeCanvas.EXIT_COMMAND ) {
				TicTacToeCanvas.this.midTacToe.quitApp();
			} else if( command == TicTacToeCanvas.INTERRUPT_THINKING_COMMAND ) {
				interruptThinking();
				repaint();
			} else if( command == TicTacToeCanvas.NEW_GAME_COMMAND ) {
				newGame();
				repaint();
			}
		}
	}

	public void keyPressed( int keyCode ) {
		if( keyCode == getKeyCode( Canvas.UP ) ) {
			midTacToe.userPlayer = 1 - midTacToe.userPlayer;
			upStatus = STATUS_LEGAL;
			repaint();
		} else if( keyCode == getKeyCode( Canvas.DOWN ) ) {
			midTacToe.userPlayer = 1 - midTacToe.userPlayer;
			downStatus = STATUS_LEGAL;
			repaint();
		} else if( keyCode == getKeyCode( Canvas.LEFT ) ) {
			midTacToe.difficulty--;
			if( midTacToe.difficulty < 0 ) {
				leftStatus = STATUS_ILLEGAL;
				midTacToe.difficulty = 0;
			} else {
				leftStatus = STATUS_LEGAL;
			}
			repaint();
		} else if( keyCode == getKeyCode( Canvas.RIGHT ) ) {
			midTacToe.difficulty++;
			if( midTacToe.difficulty > 9 ) {
				midTacToe.difficulty = 9;
				rightStatus = STATUS_ILLEGAL;
			} else {
				rightStatus = STATUS_LEGAL;
			}
			repaint();
		} else if( keyCode == Canvas.KEY_NUM1 ) {
			makeMove( 0 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM2 ) {
			makeMove( 1 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM3 ) {
			makeMove( 2 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM4 ) {
			makeMove( 3 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM5 ) {
			makeMove( 4 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM6 ) {
			makeMove( 5 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM7 ) {
			makeMove( 6 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM8 ) {
			makeMove( 7 );
			repaint();
		} else if( keyCode == Canvas.KEY_NUM9 ) {
			makeMove( 8 );
			repaint();
		}
	}

	public void keyReleased( int keyCode ) {
		if( keyCode == getKeyCode( Canvas.UP ) ) {
			upStatus = STATUS_NULL;
			repaint();
		} else if( keyCode == getKeyCode( Canvas.DOWN ) ) {
			downStatus = STATUS_NULL;
			repaint();
		} else if( keyCode == getKeyCode( Canvas.LEFT ) ) {
			leftStatus = STATUS_NULL;
			repaint();
		} else if( keyCode == getKeyCode( Canvas.RIGHT ) ) {
			rightStatus = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM1 ) {
			numbersStatus[0] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM2 ) {
			numbersStatus[1] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM3 ) {
			numbersStatus[2] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM4 ) {
			numbersStatus[3] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM5 ) {
			numbersStatus[4] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM6 ) {
			numbersStatus[5] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM7 ) {
			numbersStatus[6] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM8 ) {
			numbersStatus[7] = STATUS_NULL;
			repaint();
		} else if( keyCode == Canvas.KEY_NUM9 ) {
			numbersStatus[8] = STATUS_NULL;
			repaint();
		}
	}
}