/*
 * MIDTacToe: tic-tac-toe for modile phones
 * Copyright (C) 2003  Jason Pratt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jasonpratt.midtactoe;

/**
 * <p>Copyright: Copyright (c) 2003</p>
 * @author Jason Pratt
 */
public class TicTacToeEngine {
	public final static int X_PLAYER = 0;
	public final static int O_PLAYER = 1;

	public final static int TOP_X =          0x00000001;
	public final static int TOP_O =          0x00000002;
	public final static int TOP =            0x00000003;
	public final static int RIGHT_X =        0x00000004;
	public final static int RIGHT_O =        0x00000008;
	public final static int RIGHT =          0x0000000C;
	public final static int BOTTOM_X =       0x00000010;
	public final static int BOTTOM_O =       0x00000020;
	public final static int BOTTOM =         0x00000030;
	public final static int LEFT_X =         0x00000040;
	public final static int LEFT_O =         0x00000080;
	public final static int LEFT =           0x000000C0;
	public final static int TOP_RIGHT_X =    0x00000100;
	public final static int TOP_RIGHT_O =    0x00000200;
	public final static int TOP_RIGHT =      0x00000300;
	public final static int BOTTOM_RIGHT_X = 0x00000400;
	public final static int BOTTOM_RIGHT_O = 0x00000800;
	public final static int BOTTOM_RIGHT =   0x00000C00;
	public final static int BOTTOM_LEFT_X =  0x00001000;
	public final static int BOTTOM_LEFT_O =  0x00002000;
	public final static int BOTTOM_LEFT =    0x00003000;
	public final static int TOP_LEFT_X =     0x00004000;
	public final static int TOP_LEFT_O =     0x00008000;
	public final static int TOP_LEFT =       0x0000C000;
	public final static int CENTER_X =       0x00010000;
	public final static int CENTER_O =       0x00020000;
	public final static int CENTER =         0x00030000;

	public final static int CENTER_COLUMN = TOP | CENTER | BOTTOM;
	public final static int ROTATE_MASK_1 = TOP | RIGHT | BOTTOM | TOP_RIGHT | BOTTOM_RIGHT | BOTTOM_LEFT;
	public final static int ROTATE_MASK_2 = LEFT | TOP_LEFT;

	public final static int NULL_BOARD =     0xFFFFFFFF;

	public final static int[] X_WINS = new int[] {
		TOP_LEFT_X | TOP_X | TOP_RIGHT_X,
		LEFT_X | CENTER_X | RIGHT_X,
		BOTTOM_LEFT_X | BOTTOM_X | BOTTOM_RIGHT_X,
		TOP_LEFT_X | LEFT_X | BOTTOM_LEFT_X,
		TOP_X | CENTER_X | BOTTOM_X,
		TOP_RIGHT_X | RIGHT_X | BOTTOM_RIGHT_X,
		TOP_LEFT_X | CENTER_X | BOTTOM_RIGHT_X,
		TOP_RIGHT_X | CENTER_X | BOTTOM_LEFT_X,
	};
	public final static int[] O_WINS = new int[] {
		TOP_LEFT_O | TOP_O | TOP_RIGHT_O,
		LEFT_O | CENTER_O | RIGHT_O,
		BOTTOM_LEFT_O | BOTTOM_O | BOTTOM_RIGHT_O,
		TOP_LEFT_O | LEFT_O | BOTTOM_LEFT_O,
		TOP_O | CENTER_O | BOTTOM_O,
		TOP_RIGHT_O | RIGHT_O | BOTTOM_RIGHT_O,
		TOP_LEFT_O | CENTER_O | BOTTOM_RIGHT_O,
		TOP_RIGHT_O | CENTER_O | BOTTOM_LEFT_O,
	};

	public final static int EMPTY_CACHE_VALUE = 0xF;
	public final static int[] CACHE_MASKS = new int[] {
		0xFFFFFFF0,
		0xFFFFFF0F,
		0xFFFFF0FF,
		0xFFFF0FFF,
		0xFFF0FFFF,
		0xFF0FFFFF,
		0xF0FFFFFF,
		0x0FFFFFFF
	};

	public final static java.util.Random rand = new java.util.Random();

	public final static int[] xMoves = new int[] { TOP_LEFT_X, TOP_X, TOP_RIGHT_X,  LEFT_X, CENTER_X, RIGHT_X,   BOTTOM_LEFT_X, BOTTOM_X, BOTTOM_RIGHT_X };
	public final static int[] oMoves = new int[] { TOP_LEFT_O, TOP_O, TOP_RIGHT_O,  LEFT_O, CENTER_O, RIGHT_O,   BOTTOM_LEFT_O, BOTTOM_O, BOTTOM_RIGHT_O };
	public final static int[] allMoves = new int[] { TOP_LEFT, TOP, TOP_RIGHT,  LEFT, CENTER, RIGHT,   BOTTOM_LEFT, BOTTOM, BOTTOM_RIGHT };
	protected static int[] moveOrder = new int[] { 0, 1, 2, 3, 4, 5, 6, 7, 8 };
	protected static int[] tempMoves = new int[9];

	protected static int[] boardValueCache = new int[0x8000];
	protected static int[] clearedCache = new int[0x8000];

	protected static int boardCount;

	protected static boolean interrupted;

	public static void interruptCalculation() {
		interrupted = true;
	}

	public static int getBestMove( int board, int player, int depth ) {
		boardCount = 0;
		clearBoardValueCache();
		randomizeMoveOrder( 9 );
		int[] newBoards = generateLegalMoves( board, player );
		int size = newBoards.length;
		int value;
		int bestMove = 0;

		if( player == X_PLAYER ) {
			int max = Integer.MIN_VALUE;
			for( int i = 0; i < size; i++ ) {
				value = getCachedValueForBoard( newBoards[i] );
				if( value == EMPTY_CACHE_VALUE ) {
					value = minimax( newBoards[i], 1 - player, depth - 1 /*, Integer.MIN_VALUE, Integer.MAX_VALUE*/ );
					putCachedValueForBoardAndSymmetries( newBoards[i], value );
				}
//				printBoard( newBoards[i] );
//				System.out.println( "value: " + value );
//				System.out.println( "-----------------" );
				if( value > max ) {
					max = value;
					bestMove = newBoards[i];
				}
			}
//			System.out.println( boardCount + " boards examined." );
			interrupted = false;
			return bestMove;
		} else {
			int min = Integer.MAX_VALUE;
			for( int i = 0; i < size; i++ ) {
				value = getCachedValueForBoard( newBoards[i] );
				if( value == EMPTY_CACHE_VALUE ) {
					value = minimax( newBoards[i], 1 - player, depth - 1 /*, Integer.MIN_VALUE, Integer.MAX_VALUE*/ );
					putCachedValueForBoardAndSymmetries( newBoards[i], value );
				}
//				printBoard( newBoards[i] );
//				System.out.println( "value: " + value );
//				System.out.println( "-----------------" );
				if( value < min ) {
					min = value;
					bestMove = newBoards[i];
				}
			}
//			System.out.println( boardCount + " boards examined." );
			interrupted = false;
			return bestMove;
		}
	}

	public static int minimax( int board, int player, int depth ) {
		if( interrupted ) {
			return 0;
		}

		boardCount++;
		if( isXWin( board ) ) {
			return getBoardValue( board );
		} else if( isOWin( board ) ) {
			return -getBoardValue( board );
		} else if( depth < 1 ) {
			return 0;
		} else {
			randomizeMoveOrder( 3 );
			int[] newBoards = generateLegalMoves( board, player );
			int size = newBoards.length;
			int result;

			if( player == X_PLAYER ) {
				int max = Integer.MIN_VALUE;
				for( int i = 0; i < size; i++ ) {
					result = getCachedValueForBoard( newBoards[i] );
					if( result == EMPTY_CACHE_VALUE ) {
						result = minimax( newBoards[i], 1 - player, depth - 1 );
						putCachedValueForBoardAndSymmetries( newBoards[i], result );
					}
					if( result > max ) {
						max = result;
					}
				}
				return max;
			} else {
				int min = Integer.MAX_VALUE;
				for( int i = 0; i < size; i++ ) {
					result = getCachedValueForBoard( newBoards[i] );
					if( result == EMPTY_CACHE_VALUE ) {
						result = minimax( newBoards[i], 1 - player, depth - 1 );
						putCachedValueForBoardAndSymmetries( newBoards[i], result );
					}
					if( result < min ) {
						min = result;
					}
				}
				return min;
			}
		}
	}

// alpha/beta doesn't work with the cache
//	public static int minimax( int board, int player, int depth, int alpha, int beta ) {
//		if( isXWin( board ) ) {
//			return getBoardValue( board );
//		} else if( isOWin( board ) ) {
//			return -getBoardValue( board );
//		} else if( depth < 1 ) {
//			return 0;
//		} else {
//			randomizeMoveOrder( 4 );
//			int[] newBoards = generateLegalNonSymmetricMoves( board, player );
//			int size = newBoards.length;
//			int result;
//
//			if( player == X_PLAYER ) {
//				for( int i = 0; i < size; i++ ) {
//					result = getCachedValueForBoard( newBoards[i] );
//					if( result == EMPTY_CACHE_VALUE ) {
//						result = minimax( newBoards[i], 1 - player, depth - 1, alpha, beta );
//						putCachedValueForBoardAndSymmetries( newBoards[i], result );
//					}
//					if( result > alpha ) {
//						alpha = result;
//					}
//					if( alpha >= beta ) {
//						break;
//					}
//				}
//				return alpha;
//			} else {
//				for( int i = 0; i < size; i++ ) {
//					result = getCachedValueForBoard( newBoards[i] );
//					if( result == EMPTY_CACHE_VALUE ) {
//						result = minimax( newBoards[i], 1 - player, depth - 1, alpha, beta );
//						putCachedValueForBoardAndSymmetries( newBoards[i], result );
//					}
//					if( result < beta ) {
//						beta = result;
//					}
//					if( alpha >= beta ) {
//						break;
//					}
//				}
//				return beta;
//			}
//		}
//	}

	public static boolean isXWin( int board ) {
		int size = X_WINS.length;
		int win;
		for( int i = 0; i < size; i++ ) {
			win = X_WINS[i];
			if( (board & win) == win ) {
				return true;
			}
		}

		return false;
	}

	public static boolean isOWin( int board ) {
		int size = O_WINS.length;
		int win;
		for( int i = 0; i < size; i++ ) {
			win = O_WINS[i];
			if( (board & win) == win ) {
				return true;
			}
		}

		return false;
	}

	public static int getXWinPositions( int board ) {
		int positions = 0;
		int size = X_WINS.length;
		int win;
		for( int i = 0; i < size; i++ ) {
			win = X_WINS[i];
			if( (board & win) == win ) {
				positions |= win;
			}
		}

		return positions;
	}

	public static int getOWinPositions( int board ) {
		int positions = 0;
		int size = O_WINS.length;
		int win;
		for( int i = 0; i < size; i++ ) {
			win = O_WINS[i];
			if( (board & win) == win ) {
				positions |= win;
			}
		}

		return positions;
	}

	public static int[] generateLegalMoves( int board, int player ) {
		int moveIndex;
		int newBoard;
		int count = 0;
		for( int i = 0; i < 9; i++ ) {
			moveIndex = moveOrder[i];
			if( (board & allMoves[moveIndex]) == 0 ) {
				if( player == X_PLAYER ) {
					newBoard = board | xMoves[moveIndex];
				} else {
					newBoard = board | oMoves[moveIndex];
				}

				tempMoves[count] = newBoard;
				count++;
			}
		}

		int[] result = new int[count];
		System.arraycopy( tempMoves, 0, result, 0, count );

		return result;
	}

	public static int[] generateLegalNonSymmetricMoves( int board, int player ) {
		int moveIndex;
		int newBoard;
		boolean symmetric;
		int count = 0;
		for( int i = 0; i < 9; i++ ) {
			moveIndex = moveOrder[i];
			if( (board & allMoves[moveIndex]) == 0 ) {
				if( player == X_PLAYER ) {
					newBoard = board | xMoves[moveIndex];
				} else {
					newBoard = board | oMoves[moveIndex];
				}

				symmetric = false;
				for( int j = 0; j < count; j++ ) {
					if( isSymmetric( newBoard, tempMoves[j] ) ) {
						symmetric = true;
						break;
					}
				}

				if( ! symmetric ) {
					tempMoves[count] = newBoard;
					count++;
				}
			}
		}

		int[] result = new int[count];
		System.arraycopy( tempMoves, 0, result, 0, count );

		return result;
	}

	public static boolean isSymmetric( int board1, int board2 ) {
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}

		board2 = flipBoardAcrossVerticalAxis( board2 );
		if( board2 == board1 ) {
			return true;
		}
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}
		board2 = rotateBoardClockwise( board2 );
		if( board2 == board1 ) {
			return true;
		}

		return false;
	}

	public static int getNumMovesLeft( int board ) {
		int movesLeft = 9;
		for( int i = 0; i < 9; i++ ) {
			if( (board & allMoves[i]) > 0 ) {
				movesLeft--;
			}
		}
		return movesLeft;
	}

	public static void randomizeMoveOrder( int howMany ) {
		int r, temp;

		for( int i = 0; i < howMany; i++ ) {
			r = i + (Math.abs( rand.nextInt() ) % (9 - i));
			if( r != i ) {
				// swap
				moveOrder[i] = moveOrder[i] ^ moveOrder[r];
				moveOrder[r] = moveOrder[i] ^ moveOrder[r];
				moveOrder[i] = moveOrder[i] ^ moveOrder[r];
			}
		}
	}

	public static int flipBoardAcrossVerticalAxis( int board ) {
		return (board & CENTER_COLUMN) |
				((board & RIGHT) << 4) |
				((board & LEFT) >> 4) |
				((board & TOP_RIGHT) << 6) |
				((board & TOP_LEFT) >> 6) |
				((board & BOTTOM_RIGHT) << 2) |
				((board & BOTTOM_LEFT) >> 2);
	}

	public static int rotateBoardClockwise( int board ) {
		return (board & CENTER) |
				((board & ROTATE_MASK_1) << 2) |
				((board & ROTATE_MASK_2) >> 6);

	}

	// board value equals number of blank spaces plus one.
	public static int getBoardValue( int board ) {
		int value = 1;
		for( int i = 0; i < 9; i++ ) {
			if( (board & allMoves[i]) == 0 ) {
				value++;
			}
		}
		return value;
	}

	//////////////////////
	// board value cache
	//////////////////////
	public static void fillClearedCache() {
		for( int i = 0; i < 0x8000; i++ ) {
			clearedCache[i] = 0xFFFFFFFF;
		}
	}

	public static void clearBoardValueCache() {
		System.arraycopy( clearedCache, 0, boardValueCache, 0, clearedCache.length );
	}

	public static int getCachedValueForBoard( int board ) {
		int cacheIndex = board >> 3;
		int subIndex = board & 7;
		int bits = boardValueCache[cacheIndex] >> (subIndex*4);
		if( (bits & EMPTY_CACHE_VALUE) == EMPTY_CACHE_VALUE ) {
			return EMPTY_CACHE_VALUE;
		}
		int signBit = bits & 8;
		int magnitude = bits & 7;
		if( signBit == 0 ) {
			return magnitude;
		} else {
			return -magnitude;
		}
	}

	// value's magnitude must be 7 or less
	public static void putCachedValueForBoard( int board, int value ) {
		int cacheIndex = board >> 3;
		int subIndex = board & 7;
		int cacheValue = boardValueCache[cacheIndex];
		int nybble;
		if( value < 0 ) {
			nybble = 8 | (-value);
		} else {
			nybble = value;
		}
		boardValueCache[cacheIndex] = (cacheValue & CACHE_MASKS[subIndex]) | (nybble << (subIndex*4));
	}

	public static void putCachedValueForBoardAndSymmetries( int board, int value ) {
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
		board = flipBoardAcrossVerticalAxis( board );
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
		board = rotateBoardClockwise( board );
		putCachedValueForBoard( board, value );
	}


	public static void printBoard( int board ) {
		if( (board & TOP_LEFT_X) == TOP_LEFT_X ) {
			System.out.print( 'X' );
		} else if( (board & TOP_LEFT_O) == TOP_LEFT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & TOP_X) == TOP_X ) {
			System.out.print( 'X' );
		} else if( (board & TOP_O) == TOP_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & TOP_RIGHT_X) == TOP_RIGHT_X ) {
			System.out.print( 'X' );
		} else if( (board & TOP_RIGHT_O) == TOP_RIGHT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		System.out.println();

		if( (board & LEFT_X) == LEFT_X ) {
			System.out.print( 'X' );
		} else if( (board & LEFT_O) == LEFT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & CENTER_X) == CENTER_X ) {
			System.out.print( 'X' );
		} else if( (board & CENTER_O) == CENTER_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & RIGHT_X) == RIGHT_X ) {
			System.out.print( 'X' );
		} else if( (board & RIGHT_O) == RIGHT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		System.out.println();

		if( (board & BOTTOM_LEFT_X) == BOTTOM_LEFT_X ) {
			System.out.print( 'X' );
		} else if( (board & BOTTOM_LEFT_O) == BOTTOM_LEFT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & BOTTOM_X) == BOTTOM_X ) {
			System.out.print( 'X' );
		} else if( (board & BOTTOM_O) == BOTTOM_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		if( (board & BOTTOM_RIGHT_X) == BOTTOM_RIGHT_X ) {
			System.out.print( 'X' );
		} else if( (board & BOTTOM_RIGHT_O) == BOTTOM_RIGHT_O ) {
			System.out.print( 'O' );
		} else {
			System.out.print( '_' );
		}
		System.out.println();
	}
}